@@ -1,22 +1,22 @@
 /* pinentry-curses.c - A secure curses dialog for PIN entry, library version
-   Copyright (C) 2002, 2015 g10 Code GmbH
-
-   This file is part of PINENTRY.
-
-   PINENTRY is free software; you can redistribute it and/or modify it
-   under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
-
-   PINENTRY is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA  */
+ * Copyright (C) 2002, 2015 g10 Code GmbH
+ *
+ * This file is part of PINENTRY.
+ *
+ * PINENTRY is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * PINENTRY is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * SPDX-License-Identifier: GPL-2.0+
+ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -94,6 +94,8 @@ struct dialog
   int pin_max;
   /* Length of PIN.  */
   int pin_len;
+  int got_input;
+  int no_echo;
 
   int ok_y;
   int ok_x;
@@ -113,6 +115,7 @@ typedef struct dialog *dialog_t;
 #ifdef HAVE_NCURSESW
 typedef wchar_t CH;
 #define STRLEN(x) wcslen (x)
+#define STRWIDTH(x) wcswidth (x, wcslen (x))
 #define ADDCH(x) addnwstr (&x, 1);
 #define CHWIDTH(x) wcwidth (x)
 #define NULLCH L'\0'
@@ -121,6 +124,7 @@ typedef wchar_t CH;
 #else
 typedef char CH;
 #define STRLEN(x) strlen (x)
+#define STRWIDTH(x) strlen (x)
 #define ADDCH(x) addch ((unsigned char) x)
 #define CHWIDTH(x) 1
 #define NULLCH '\0'
@@ -128,13 +132,14 @@ typedef char CH;
 #define SPCH ' '
 #endif
 
-/* Return the next line up to MAXLEN columns wide in START and LEN.
+/* Return the next line up to MAXWIDTH columns wide in START and LEN.
+   Return value is the width needed for the line.
    The first invocation should have 0 as *LEN.  If the line ends with
    a \n, it is a normal line that will be continued.  If it is a '\0'
    the end of the text is reached after this line.  In all other cases
    there is a forced line break.  A full line is returned and will be
    continued in the next line.  */
-static void
+static int
 collect_line (int maxwidth, CH **start_p, int *len_p)
 {
   int last_space = 0;
@@ -153,11 +158,11 @@ collect_line (int maxwidth, CH **start_p, int *len_p)
 
   while (width < maxwidth - 1 && *end != NULLCH && *end != NLCH)
     {
-      len++;
-      end++;
       if (*end == SPCH)
 	last_space = len;
       width += CHWIDTH (*end);
+      len++;
+      end++;
     }
 
   if (*end != NULLCH && *end != NLCH && last_space != 0)
@@ -169,6 +174,7 @@ collect_line (int maxwidth, CH **start_p, int *len_p)
       (*start_p)[len] = NLCH;
     }
   *len_p = len + 1;
+  return width;
 }
 
 #ifdef HAVE_NCURSESW
@@ -254,6 +260,7 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
 	  {								\
 	    err = 1;							\
             pinentry->specific_err = gpg_error (GPG_ERR_LOCALE_PROBLEM); \
+            pinentry->specific_err_loc = "dialog_create_copy";          \
 	    goto out;							\
 	  }								\
       }									\
@@ -286,6 +293,7 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
 	    {								\
 	      err = 1;							\
               pinentry->specific_err = gpg_error_from_syserror ();	\
+              pinentry->specific_err_loc = "dialog_create_mk_button";   \
 	      goto out;							\
 	    }								\
 									\
@@ -311,6 +319,7 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
         {								\
 	  err = 1;							\
           pinentry->specific_err = gpg_error (GPG_ERR_LOCALE_PROBLEM);	\
+          pinentry->specific_err_loc = "dialog_create_utf8conv";        \
 	  goto out;							\
 	}								\
     }									\
@@ -337,9 +346,10 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
 
       do
 	{
-	  collect_line (size_x - 4, &start, &len);
-	  if (len > description_x)
-	    description_x = len;
+	  int width = collect_line (size_x - 4, &start, &len);
+
+	  if (width > description_x)
+	    description_x = width;
 	  y++;
 	}
       while (start[len - 1]);
@@ -378,6 +388,7 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
       err = 1;
       pinentry->specific_err = gpg_error (size_y < 0? GPG_ERR_MISSING_ENVVAR
                                           /* */     : GPG_ERR_WINDOW_TOO_SMALL);
+      pinentry->specific_err_loc = "dialog_create";
       goto out;
     }
 
@@ -408,7 +419,7 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
       new_x = MIN_PINENTRY_LENGTH;
       if (prompt)
 	{
-	  new_x += STRLEN (prompt) + 1;	/* One space after prompt.  */
+	  new_x += STRWIDTH (prompt) + 1;	/* One space after prompt.  */
 	}
       if (new_x > size_x - 4)
 	new_x = size_x - 4;
@@ -434,6 +445,7 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
       err = 1;
       pinentry->specific_err = gpg_error (size_x < 0? GPG_ERR_MISSING_ENVVAR
                                           /* */     : GPG_ERR_WINDOW_TOO_SMALL);
+      pinentry->specific_err_loc = "dialog_create";
       goto out;
     }
 
@@ -536,11 +548,12 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
       if (prompt)
 	{
 	  CH *p = prompt;
-	  i = STRLEN (prompt);
+	  i = STRWIDTH (prompt);
 	  if (i > x - 4 - MIN_PINENTRY_LENGTH)
 	    i = x - 4 - MIN_PINENTRY_LENGTH;
 	  dialog->pin_x += i + 1;
 	  dialog->pin_size -= i + 1;
+	  i = STRLEN (prompt);
 	  while (i-- > 0)
 	    {
 	      ADDCH (*(p++));
@@ -591,6 +604,9 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
       addstr (dialog->ok);
     }
 
+  dialog->got_input = 0;
+  dialog->no_echo = 0;
+
  out:
   if (description)
     free (description);
@@ -725,6 +741,12 @@ dialog_input (dialog_t diag, int alt, int chr)
 		diag->pin_loc = diag->pin_len;
 	    }
 	}
+      else if (!diag->got_input)
+	{
+	  diag->no_echo = 1;
+	  move (diag->pin_y, diag->pin_x);
+	  addstr ("[no echo]");
+	}
       break;
 
     case 'l' - 'a' + 1: /* control-l */
@@ -796,24 +818,30 @@ dialog_input (dialog_t diag, int alt, int chr)
       break;
     }
 
-  if (old_loc < diag->pin_loc)
-    {
-      move (diag->pin_y, diag->pin_x + old_loc);
-      while (old_loc++ < diag->pin_loc)
-	addch ('*');
-    }
-  else if (old_loc > diag->pin_loc)
+  diag->got_input = 1;
+
+  if (!diag->no_echo)
     {
+      if (old_loc < diag->pin_loc)
+	{
+	  move (diag->pin_y, diag->pin_x + old_loc);
+	  while (old_loc++ < diag->pin_loc)
+	    addch ('*');
+	}
+      else if (old_loc > diag->pin_loc)
+	{
+	  move (diag->pin_y, diag->pin_x + diag->pin_loc);
+	  while (old_loc-- > diag->pin_loc)
+	    addch ('_');
+	}
       move (diag->pin_y, diag->pin_x + diag->pin_loc);
-      while (old_loc-- > diag->pin_loc)
-	addch ('_');
     }
-  move (diag->pin_y, diag->pin_x + diag->pin_loc);
 }
 
 static int
 dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
 {
+  int confirm_mode = !pinentry->pin;
   struct dialog diag;
   FILE *ttyfi = NULL;
   FILE *ttyfo = NULL;
@@ -824,6 +852,7 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
 #ifndef HAVE_DOSISH_SYSTEM
   int no_input = 1;
 #endif
+
 #ifdef HAVE_NCURSESW
   char *old_ctype = NULL;
 
@@ -843,7 +872,8 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
       if (!ttyfi)
         {
           pinentry->specific_err = gpg_error_from_syserror ();
-          return -1;
+          pinentry->specific_err_loc = "open_tty_for_read";
+          return confirm_mode? 0 : -1;
         }
       ttyfo = fopen (tty_name, "w");
       if (!ttyfo)
@@ -852,7 +882,8 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
 	  fclose (ttyfi);
 	  errno = err;
           pinentry->specific_err = gpg_error_from_syserror ();
-	  return -1;
+          pinentry->specific_err_loc = "open_tty_for_write";
+	  return confirm_mode? 0 : -1;
 	}
       screen = newterm (tty_type, ttyfo, ttyfi);
       set_term (screen);
@@ -865,7 +896,8 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
             {
               errno = ENOTTY;
               pinentry->specific_err = gpg_error_from_syserror ();
-              return -1;
+              pinentry->specific_err_loc = "isatty";
+              return confirm_mode? 0 : -1;
             }
 	  init_screen = 1;
 	  initscr ();
@@ -879,6 +911,14 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
   cbreak ();		/* Take input chars one at a time, no wait for \n.  */
   noecho ();		/* Don't echo input - in color.  */
 
+  if (pinentry->ttyalert)
+    {
+      if (! strcmp(pinentry->ttyalert, "beep"))
+	beep ();
+      else if (! strcmp(pinentry->ttyalert, "flash"))
+	flash ();
+    }
+
   if (has_colors ())
     {
       start_color ();
@@ -925,8 +965,7 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
         fclose (ttyfo);
       return -2;
     }
-  dialog_switch_pos (&diag,
-		     diag.pinentry->pin ? DIALOG_POS_PIN : DIALOG_POS_OK);
+  dialog_switch_pos (&diag, confirm_mode? DIALOG_POS_OK : DIALOG_POS_PIN);
 
 #ifndef HAVE_DOSISH_SYSTEM
   wtimeout (stdscr, 70);
@@ -966,7 +1005,7 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
 	  switch (diag.pos)
 	    {
 	    case DIALOG_POS_OK:
-	      if (diag.pinentry->pin)
+	      if (!confirm_mode)
 		dialog_switch_pos (&diag, DIALOG_POS_PIN);
 	      break;
 	    case DIALOG_POS_NOTOK:
@@ -1020,10 +1059,10 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
 	      dialog_switch_pos (&diag, DIALOG_POS_CANCEL);
 	      break;
 	    case DIALOG_POS_CANCEL:
-	      if (diag.pinentry->pin)
-		dialog_switch_pos (&diag, DIALOG_POS_PIN);
-	      else
+	      if (confirm_mode)
 		dialog_switch_pos (&diag, DIALOG_POS_OK);
+	      else
+		dialog_switch_pos (&diag, DIALOG_POS_PIN);
 	      break;
 	    default:
 	      break;
@@ -1064,10 +1103,12 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
     }
   while (!done);
 
-  if (diag.pinentry->pin)
-    /* NUL terminate the passphrase.  dialog_run makes sure there is
-       enough space for the terminating NUL byte.  */
-    diag.pinentry->pin[diag.pin_len] = 0;
+  if (!confirm_mode)
+    {
+      /* NUL terminate the passphrase.  dialog_run makes sure there is
+         enough space for the terminating NUL byte.  */
+      diag.pinentry->pin[diag.pin_len] = 0;
+    }
 
   set_cursor_state (1);
   endwin ();
@@ -1092,7 +1133,7 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
   if (diag.notok)
     free (diag.notok);
 
-  if (pinentry->pin)
+  if (!confirm_mode)
     {
       pinentry->locale_err = 1;
       pin_utf8 = pinentry_local_to_utf8 (pinentry->lc_ctype, pinentry->pin, 1);
@@ -1109,10 +1150,11 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
   if (done == -2)
     pinentry->canceled = 1;
 
-  if (diag.pinentry->pin)
-    return done < 0 ? -1 : diag.pin_len;
-  else
+  /* In confirm mode return cancel instead of error.  */
+  if (confirm_mode)
     return done < 0 ? 0 : 1;
+
+  return done < 0 ? -1 : diag.pin_len;
 }
 
 
@@ -1168,7 +1210,7 @@ curses_cmd_handler (pinentry_t pinentry)
     }
 #endif
 
-  rc = dialog_run (pinentry, pinentry->ttyname, pinentry->ttytype);
+  rc = dialog_run (pinentry, pinentry->ttyname, pinentry->ttytype_l);
   do_touch_file (pinentry);
   return rc;
 }
