@@ -1,21 +1,22 @@
 /* pinentry-emacs.c - A secure emacs dialog for PIN entry, library version
-   Copyright (C) 2015 Daiki Ueno
-
-   This file is part of PINENTRY.
-
-   PINENTRY is free software; you can redistribute it and/or modify it
-   under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
-
-   PINENTRY is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, see <https://www.gnu.org/licenses/>.
-*/
+ * Copyright (C) 2015 Daiki Ueno
+ *
+ * This file is part of PINENTRY.
+ *
+ * PINENTRY is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * PINENTRY is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
+ * SPDX-License-Identifier: GPL-2.0+
+ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -467,8 +468,14 @@ set_label (pinentry_t pe, const char *name, const char *value)
 static void
 set_labels (pinentry_t pe)
 {
-  if (pe->title)
-    set_label (pe, "SETTITLE", pe->title);
+  char *p;
+
+  p = pinentry_get_title (pe);
+  if (p)
+    {
+      set_label (pe, "SETTITLE", p);
+      free (p);
+    }
   if (pe->description)
     set_label (pe, "SETDESC", pe->description);
   if (pe->error)
@@ -644,7 +651,10 @@ initial_emacs_cmd_handler (pinentry_t pe)
   if (emacs_socket < 0)
     pinentry_cmd_handler = fallback_cmd_handler;
   else
-    pinentry_cmd_handler = emacs_cmd_handler;
+    {
+      pinentry_cmd_handler = emacs_cmd_handler;
+      pinentry_set_flavor_flag ("emacs");
+    }
 
   return (* pinentry_cmd_handler) (pe);
 }
@@ -670,6 +680,22 @@ pinentry_enable_emacs_cmd_handler (void)
   pinentry_cmd_handler = initial_emacs_cmd_handler;
 }
 
+
+/* Returns true if the Emacs pinentry is enabled.  The value is 1
+ * before the first connection with Emacs has been done and 2 if the
+ * connection to Emacs has been establish.  Returns false if the Emacs
+ * pinentry is not enabled.  */
+int
+pinentry_emacs_status (void)
+{
+  if (pinentry_cmd_handler == initial_emacs_cmd_handler)
+    return 1;
+  else if (pinentry_cmd_handler == emacs_cmd_handler)
+    return 2;
+  else
+    return 0;
+}
+
 int
 pinentry_emacs_init (void)
 {
