@@ -1,27 +1,26 @@
 /* main.cpp - A Qt dialog for PIN entry.
-
-   Copyright (C) 2002, 2008 Klarälvdalens Datakonsult AB (KDAB)
-   Copyright (C) 2003 g10 Code GmbH
-   Copyright 2007 Ingo Klöcker
-
-   Written by Steffen Hansen <steffen@klaralvdalens-datakonsult.se>.
-   Modified by Marcus Brinkmann <marcus@g10code.de>.
-   Modified by Marc Mutz <marc@kdab.com>
-
-   This program is free software; you can redistribute it and/or
-   modify it under the terms of the GNU General Public License as
-   published by the Free Software Foundation; either version 2 of the
-   License, or (at your option) any later version.
-
-   This program is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
-*/
+ * Copyright (C) 2002, 2008 Klarälvdalens Datakonsult AB (KDAB)
+ * Copyright (C) 2003 g10 Code GmbH
+ * Copyright 2007 Ingo Klöcker
+ *
+ * Written by Steffen Hansen <steffen@klaralvdalens-datakonsult.se>.
+ * Modified by Marcus Brinkmann <marcus@g10code.de>.
+ * Modified by Marc Mutz <marc@kdab.com>
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
+ * SPDX-License-Identifier: GPL-2.0+
+ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -31,18 +30,19 @@
 #include "pinentrydialog.h"
 #include "pinentry.h"
 
-#include <qapplication.h>
+#include <QApplication>
+#include <QDebug>
 #include <QIcon>
-#include <QString>
-#include <qwidget.h>
-#include <qmessagebox.h>
+#include <QMessageBox>
 #include <QPushButton>
+#include <QString>
+#include <QWidget>
+#include <QWindow>
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
 
-#include <memory>
 #include <stdexcept>
 #include <gpg-error.h>
 
@@ -63,6 +63,10 @@
   #endif
 #endif
 
+#ifdef Q_OS_WIN
+#include <windows.h>
+#endif
+
 static QString escape_accel(const QString &s)
 {
 
@@ -99,22 +103,6 @@ static QString escape_accel(const QString &s)
     return result;
 }
 
-/* Hack for creating a QWidget with a "foreign" window ID */
-class ForeignWidget : public QWidget
-{
-public:
-    explicit ForeignWidget(WId wid) : QWidget(0)
-    {
-        QWidget::destroy();
-        create(wid, false, false);
-    }
-
-    ~ForeignWidget()
-    {
-        destroy(false, false);
-    }
-};
-
 namespace
 {
 class InvalidUtf8 : public std::invalid_argument
@@ -141,15 +129,43 @@ static QString from_utf8(const char *s)
     return result;
 }
 
+static void
+setup_foreground_window(QWidget *widget, WId parentWid)
+{
+    /* For windows set the desktop window as the transient parent */
+    QWindow *parentWindow = nullptr;
+    if (parentWid) {
+        parentWindow = QWindow::fromWinId(parentWid);
+    }
+#ifdef Q_OS_WIN
+    if (!parentWindow) {
+        HWND desktop = GetDesktopWindow();
+        if (desktop) {
+            parentWindow = QWindow::fromWinId((WId) desktop);
+        }
+    }
+#endif
+    if (parentWindow) {
+        // Ensure that we have a native wid
+        widget->winId();
+        QWindow *wndHandle = widget->windowHandle();
+
+        if (wndHandle) {
+            wndHandle->setTransientParent(parentWindow);
+        }
+    }
+    widget->setWindowFlags(Qt::Window |
+                           Qt::CustomizeWindowHint |
+                           Qt::WindowTitleHint |
+                           Qt::WindowCloseButtonHint |
+                           Qt::WindowStaysOnTopHint |
+                           Qt::WindowMinimizeButtonHint);
+}
+
 static int
 qt_cmd_handler(pinentry_t pe)
 {
-    QWidget *parent = 0;
-
-    /* FIXME: Add parent window ID to pinentry and GTK.  */
-    if (pe->parent_wid) {
-        parent = new ForeignWidget((WId) pe->parent_wid);
-    }
+    char *str;
 
     int want_pass = !!pe->pin;
 
@@ -161,9 +177,13 @@ qt_cmd_handler(pinentry_t pe)
         pe->cancel         ? escape_accel(from_utf8(pe->cancel)) :
         pe->default_cancel ? escape_accel(from_utf8(pe->default_cancel)) :
         /* else */           QLatin1String("&Cancel") ;
+
+    str = pinentry_get_title (pe);
     const QString title =
-        pe->title ? from_utf8(pe->title) :
+        str       ? from_utf8(str) :
         /* else */  QLatin1String("pinentry-qt") ;
+    free (str);
+
     const QString repeatError =
         pe->repeat_error_string ? from_utf8(pe->repeat_error_string) :
                                   QLatin1String("Passphrases do not match");
@@ -177,17 +197,29 @@ qt_cmd_handler(pinentry_t pe)
         pe->default_tt_hide ? from_utf8(pe->default_tt_hide) :
                               QLatin1String("Hide passphrase");
 
+    const QString generateLbl = pe->genpin_label ? from_utf8(pe->genpin_label) :
+                                QString();
+    const QString generateTT = pe->genpin_tt ? from_utf8(pe->genpin_tt) :
+                               QString();
+
 
     if (want_pass) {
-        PinEntryDialog pinentry(parent, 0, pe->timeout, true, !!pe->quality_bar,
-                                repeatString, visibilityTT, hideTT);
+        char *str;
 
+        PinEntryDialog pinentry(nullptr, 0, pe->timeout, true, !!pe->quality_bar,
+                                repeatString, visibilityTT, hideTT);
+        setup_foreground_window(&pinentry, pe->parent_wid);
         pinentry.setPinentryInfo(pe);
         pinentry.setPrompt(escape_accel(from_utf8(pe->prompt)));
         pinentry.setDescription(from_utf8(pe->description));
         pinentry.setRepeatErrorText(repeatError);
-        if (pe->title) {
-            pinentry.setWindowTitle(from_utf8(pe->title));
+        pinentry.setGenpinLabel(generateLbl);
+        pinentry.setGenpinTT(generateTT);
+
+        str = pinentry_get_title (pe);
+        if (str) {
+            pinentry.setWindowTitle(from_utf8(str));
+            free (str);
         }
 
         /* If we reuse the same dialog window.  */
@@ -239,7 +271,8 @@ qt_cmd_handler(pinentry_t pe)
             pe->notok      ? QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel :
             /* else */       QMessageBox::Ok | QMessageBox::Cancel ;
 
-        PinentryConfirm box(QMessageBox::Information, pe->timeout, title, desc, buttons, parent);
+        PinentryConfirm box(QMessageBox::Information, pe->timeout, title, desc, buttons, nullptr);
+        setup_foreground_window(&box, pe->parent_wid);
 
         const struct {
             QMessageBox::StandardButton button;
@@ -303,11 +336,13 @@ main(int argc, char *argv[])
 {
     pinentry_init("pinentry-qt");
 
-    std::auto_ptr<QApplication> app;
+    QApplication *app = NULL;
+    int new_argc = 0;
 
 #ifdef FALLBACK_CURSES
     if (!pinentry_have_display(argc, argv)) {
         pinentry_cmd_handler = curses_cmd_handler;
+        pinentry_set_flavor_flag ("curses");
     } else
 #endif
     {
@@ -342,14 +377,21 @@ main(int argc, char *argv[])
                 p += strlen(argv[i]) + 1;
             }
 
-        /* We use a modal dialog window, so we don't need the application
-           window anymore.  */
-        i = argc;
-        app.reset(new QApplication(i, new_argv));
+        /* Note: QApplication uses int &argc so argc has to be valid
+         * for the full lifetime of the application.
+         *
+         * As Qt might modify argc / argv we use copies here so that
+         * we do not loose options that are handled in both. e.g. display.
+         */
+        new_argc = argc;
+        Q_ASSERT (new_argc);
+        app = new QApplication(new_argc, new_argv);
         app->setWindowIcon(QIcon(QLatin1String(":/document-encrypt.png")));
     }
 
     pinentry_parse_opts(argc, argv);
 
-    return pinentry_loop() ? EXIT_FAILURE : EXIT_SUCCESS ;
+    int rc = pinentry_loop();
+    delete app;
+    return rc ? EXIT_FAILURE : EXIT_SUCCESS ;
 }
